import functools
import logging
import time
from typing import List, Optional, Iterator, Callable

import opik
import opik.logging_messages as logging_messages
import opik.opik_context as opik_context
from opik.api_objects import opik_client, trace, local_recording
from opik.api_objects.dataset import dataset_item
from opik.api_objects.experiment import experiment
from opik.api_objects.dataset import execution_policy as dataset_execution_policy
from opik.evaluation import rest_operations, test_case, test_result
from opik.evaluation.types import LLMTask, ScoringKeyMappingType
from opik.message_processing.emulation import models
from opik.types import TraceSource

from . import evaluation_tasks_executor, exception_analyzer, helpers, metrics_evaluator
from .types import EvaluationTask
from ..metrics import base_metric, score_result


LOGGER = logging.getLogger(__name__)

EVALUATION_TASK_NAME = "evaluation_task"


def get_item_execution_policy(
    item: dataset_item.DatasetItem,
    default_policy: dataset_execution_policy.ExecutionPolicy,
) -> dataset_execution_policy.ExecutionPolicy:
    """
    Get execution policy for a dataset item.

    If the item has its own execution policy, merge it with the default.
    Item-level values override default values.

    Args:
        item: The dataset item.
        default_policy: Default execution policy from suite level.

    Returns:
        Merged execution policy for this item.
    """
    if item.execution_policy is None:
        return default_policy

    return {
        "runs_per_item": (
            item.execution_policy.runs_per_item
            if item.execution_policy.runs_per_item is not None
            else default_policy.get("runs_per_item", 1)
        ),
        "pass_threshold": (
            item.execution_policy.pass_threshold
            if item.execution_policy.pass_threshold is not None
            else default_policy.get("pass_threshold", 1)
        ),
    }


class EvaluationEngine:
    """
    Stateless evaluation executor.

    Only stores configuration (client, workers, verbosity).
    All flow-specific data (metrics, key mappings) is passed as method parameters.
    """

    def __init__(
        self,
        client: opik_client.Opik,
        project_name: Optional[str],
        workers: int,
        verbose: int,
        source: TraceSource,
    ) -> None:
        self._client = client
        self._project_name = project_name
        self._workers = workers
        self._verbose = verbose
        self._source = source

    # --- Private: metrics & scoring ---

    @opik.track(  # type: ignore[attr-defined,has-type]
        name="metrics_calculation",
        ignore_arguments=[
            "regular_metrics",
            "scoring_key_mapping",
            "evaluator_model",
        ],
    )
    def _compute_test_result_for_test_case(
        self,
        test_case_: test_case.TestCase,
        regular_metrics: List[base_metric.BaseMetric],
        scoring_key_mapping: ScoringKeyMappingType,
        evaluator_model: Optional[str],
        trial_id: int = 0,
    ) -> test_result.TestResult:
        item_evaluator = metrics_evaluator.build_metrics_evaluator(
            item=test_case_.dataset_item,
            regular_metrics=regular_metrics,
            scoring_key_mapping=scoring_key_mapping,
            evaluator_model=evaluator_model,
        )
        score_results, mapped_scoring_inputs = item_evaluator.compute_regular_scores(
            dataset_item_content=test_case_.dataset_item_content,
            task_output=test_case_.task_output,
        )
        test_case_.mapped_scoring_inputs = mapped_scoring_inputs

        test_result_ = test_result.TestResult(
            test_case=test_case_,
            score_results=score_results,
            trial_id=trial_id,
        )
        rest_operations.log_test_result_feedback_scores(
            client=self._client,
            score_results=score_results,
            trace_id=test_case_.trace_id,
            project_name=self._project_name,
        )
        return test_result_

    @opik.track(  # type: ignore[attr-defined,has-type]
        name="task_span_metrics_calculation",
        ignore_arguments=["test_case_", "task_span_evaluator"],
    )
    def _compute_scores_for_test_case_with_task_span(
        self,
        trace_id: str,
        task_span: models.SpanModel,
        test_case_: test_case.TestCase,
        task_span_evaluator: metrics_evaluator.MetricsEvaluator,
    ) -> List[score_result.ScoreResult]:
        score_results, mapped_scoring_inputs = (
            task_span_evaluator.compute_task_span_scores(
                dataset_item_content=test_case_.dataset_item_content,
                task_output=test_case_.task_output,
                task_span=task_span,
            )
        )
        test_case_.mapped_scoring_inputs = mapped_scoring_inputs

        # log feedback scores
        rest_operations.log_test_result_feedback_scores(
            client=self._client,
            score_results=score_results,
            trace_id=trace_id,
            project_name=self._project_name,
        )
        return score_results

    # --- Private: task execution ---

    def _compute_test_result_for_llm_task(
        self,
        item: dataset_item.DatasetItem,
        task: LLMTask,
        trial_id: int,
        experiment_: Optional[experiment.Experiment],
        regular_metrics: List[base_metric.BaseMetric],
        scoring_key_mapping: ScoringKeyMappingType,
        evaluator_model: Optional[str],
    ) -> test_result.TestResult:
        if not hasattr(task, "opik_tracked"):
            name = task.__name__ if hasattr(task, "__name__") else "llm_task"
            task = opik.track(name=name, source=self._source)(task)  # type: ignore[attr-defined,has-type]

        item_content = item.get_content(include_id=True)
        trace_data = trace.TraceData(
            input=item_content,
            name=EVALUATION_TASK_NAME,
            created_by="evaluation",
            project_name=self._project_name,
            source=self._source,
        )

        execution_policy_dict = None
        if item.execution_policy is not None:
            execution_policy_dict = item.execution_policy.model_dump(exclude_none=True)

        with helpers.evaluate_llm_task_context(
            experiment=experiment_,
            dataset_item_id=item.id,
            trace_data=trace_data,
            client=self._client,
            execution_policy=execution_policy_dict or None,
        ):
            LOGGER.debug("[engine] Task started for item %s", item.id)
            task_start = time.perf_counter()
            try:
                task_output_ = task(item_content)
            except Exception as exception:
                LOGGER.error("[engine] Task failed for item %s: %s", item.id, exception)
                if exception_analyzer.is_llm_provider_rate_limit_error(exception):
                    LOGGER.error(
                        logging_messages.LLM_PROVIDER_RATE_LIMIT_ERROR_DETECTED_IN_EVALUATE_FUNCTION
                    )

                raise
            task_execution_time = time.perf_counter() - task_start
            LOGGER.debug(
                "[engine] Task done for item %s in %.1fs", item.id, task_execution_time
            )

            opik_context.update_current_trace(output=task_output_)

            test_case_ = test_case.TestCase(
                trace_id=trace_data.id,
                dataset_item_id=item.id,
                task_output=task_output_,
                dataset_item_content=item_content,
                dataset_item=item,
            )
            LOGGER.debug("[engine] Scoring started for item %s", item.id)
            scoring_start = time.perf_counter()
            test_result_ = self._compute_test_result_for_test_case(
                test_case_=test_case_,
                regular_metrics=regular_metrics,
                scoring_key_mapping=scoring_key_mapping,
                evaluator_model=evaluator_model,
                trial_id=trial_id,
            )
            test_result_.task_execution_time = task_execution_time
            test_result_.scoring_time = time.perf_counter() - scoring_start
            LOGGER.debug(
                "[engine] Scoring done for item %s in %.1fs",
                item.id,
                test_result_.scoring_time,
            )

        return test_result_

    # --- Private: parallel execution ---

    def _compute_test_results_with_execution_policy(
        self,
        dataset_items: Iterator[dataset_item.DatasetItem],
        task: LLMTask,
        experiment_: Optional[experiment.Experiment],
        regular_metrics: List[base_metric.BaseMetric],
        scoring_key_mapping: ScoringKeyMappingType,
        evaluator_model: Optional[str],
        description: str,
        total_items: Optional[int],
        default_execution_policy: dataset_execution_policy.ExecutionPolicy,
        show_scores_in_progress_bar: bool,
    ) -> List[test_result.TestResult]:
        """
        Execute tasks with full parallelism and item-based progress.

        Uses StreamingExecutor with group_id to track progress per item
        (not per run), so that runs_per_item > 1 shows correct item count.
        """
        with evaluation_tasks_executor.StreamingExecutor[test_result.TestResult](
            workers=self._workers,
            verbose=self._verbose,
            desc=description,
            total=total_items,
            show_score_postfix=show_scores_in_progress_bar,
        ) as executor:
            for item in dataset_items:
                item_policy = get_item_execution_policy(item, default_execution_policy)
                item_runs = item_policy.get("runs_per_item", 1)

                # Store resolved execution policy
                item.execution_policy = dataset_item.ExecutionPolicyItem(
                    runs_per_item=item_runs,
                    pass_threshold=item_policy.get("pass_threshold", 1),
                )

                # Declare group size before submitting to avoid race with callbacks
                executor.set_group_size(item.id, item_runs)

                # Submit all runs for this item
                for run_id in range(item_runs):
                    executor.submit(
                        functools.partial(
                            self._compute_test_result_for_llm_task,
                            item=item,
                            task=task,
                            trial_id=run_id,
                            experiment_=experiment_,
                            regular_metrics=regular_metrics,
                            scoring_key_mapping=scoring_key_mapping,
                            evaluator_model=evaluator_model,
                        ),
                        group_id=item.id,
                    )

            return executor.get_results()

    # --- Private: task-span metrics ---

    def _update_test_result_with_task_span_metrics(
        self,
        evaluation_task_result: test_result.TestResult,
        trace_trees: List[models.TraceModel],
        task_span_evaluator: metrics_evaluator.MetricsEvaluator,
    ) -> test_result.TestResult:
        # find related trace
        trace_id = evaluation_task_result.test_case.trace_id
        task_trace = None
        for trace_ in trace_trees:
            if trace_.id == trace_id:
                task_trace = trace_
                break

        if task_trace is None:
            raise ValueError(
                f"No trace found for test result: {evaluation_task_result}"
            )

        # find evaluation span
        if len(task_trace.spans) == 0:
            raise ValueError(
                f"Task trace contains no spans. Task span metrics require at least one span to be present in the execution trace. Test result: {evaluation_task_result}"
            )
        # the first span is the evaluation span
        evaluation_span = task_trace.spans[0]

        with helpers.evaluate_llm_task_result_spans_context(
            trace_data=trace.TraceData(
                id=trace_id,
                name=task_trace.name,
                start_time=task_trace.start_time,
                metadata=task_trace.metadata,
                input=task_trace.input,
                output=task_trace.output,
                tags=task_trace.tags,
                project_name=self._project_name,
                created_by="evaluation",
                error_info=task_trace.error_info,
                thread_id=task_trace.thread_id,
                source=self._source,
            ),
            client=self._client,
        ):
            score_results = self._compute_scores_for_test_case_with_task_span(
                trace_id=trace_id,
                task_span=evaluation_span,
                test_case_=evaluation_task_result.test_case,
                task_span_evaluator=task_span_evaluator,
            )
            # append scores to the input test result
            evaluation_task_result.score_results += score_results
            return evaluation_task_result

    def _update_test_results_with_task_span_metrics(
        self,
        test_results: List[test_result.TestResult],
        recording: local_recording._LocalRecordingHandle,
        task_span_evaluator: metrics_evaluator.MetricsEvaluator,
    ) -> None:
        """Evaluate task spans from a local recording."""
        # Get trace trees from the recording (this flushes automatically)
        trace_trees = recording.trace_trees
        if len(trace_trees) == 0:
            LOGGER.warning("No trace trees found in the local recording.")
            return

        # Create span evaluation tasks from LLM tasks evaluation results and evaluate them in parallel
        span_evaluation_tasks: List[EvaluationTask[test_result.TestResult]] = [
            functools.partial(
                self._update_test_result_with_task_span_metrics,
                evaluation_task_result=test_result_,
                trace_trees=trace_trees,
                task_span_evaluator=task_span_evaluator,
            )
            for test_result_ in test_results
        ]

        evaluation_tasks_executor.execute(
            evaluation_tasks=span_evaluation_tasks,
            workers=self._workers,
            verbose=self._verbose,
            desc="LLM task spans evaluation",
        )

        LOGGER.debug(
            "Task evaluation span handling is disabled — the evaluation has been completed."
        )

    # --- Private: shared execution with optional task-span scoring ---

    def _execute_evaluation(
        self,
        dataset_items: Iterator[dataset_item.DatasetItem],
        task: LLMTask,
        experiment_: Optional[experiment.Experiment],
        regular_metrics: List[base_metric.BaseMetric],
        task_span_metrics: List[base_metric.BaseMetric],
        scoring_key_mapping: ScoringKeyMappingType,
        evaluator_model: Optional[str],
        description: str,
        total_items: Optional[int],
        default_execution_policy: dataset_execution_policy.ExecutionPolicy,
        show_scores_in_progress_bar: bool,
    ) -> List[test_result.TestResult]:
        """
        Shared execution logic. Runs the task, scores with regular metrics,
        and optionally scores with task-span metrics.
        """
        compute: Callable[[], List[test_result.TestResult]] = functools.partial(
            self._compute_test_results_with_execution_policy,
            dataset_items=dataset_items,
            task=task,
            experiment_=experiment_,
            regular_metrics=regular_metrics,
            scoring_key_mapping=scoring_key_mapping,
            evaluator_model=evaluator_model,
            description=description,
            total_items=total_items,
            default_execution_policy=default_execution_policy,
            show_scores_in_progress_bar=show_scores_in_progress_bar,
        )

        if not task_span_metrics:
            return compute()

        LOGGER.debug(
            "Detected %d LLM task span scoring metrics — enabling handling of the LLM task evaluation span.",
            len(task_span_metrics),
        )

        task_span_evaluator = metrics_evaluator.MetricsEvaluator(
            scoring_metrics=task_span_metrics,
            scoring_key_mapping=scoring_key_mapping,
        )

        with local_recording.record_traces_locally(client=self._client) as recording:
            test_results = compute()
            self._update_test_results_with_task_span_metrics(
                test_results=test_results,
                recording=recording,
                task_span_evaluator=task_span_evaluator,
            )

        return test_results

    # --- Public API ---

    def run_and_score(
        self,
        dataset_items: Iterator[dataset_item.DatasetItem],
        task: LLMTask,
        scoring_metrics: List[base_metric.BaseMetric],
        scoring_key_mapping: Optional[ScoringKeyMappingType],
        evaluator_model: Optional[str],
        experiment_: Optional[experiment.Experiment],
        default_execution_policy: dataset_execution_policy.ExecutionPolicy,
        total_items: Optional[int],
        description: str = "Evaluation",
        show_scores_in_progress_bar: bool = True,
    ) -> List[test_result.TestResult]:
        """
        Run a task on dataset items in parallel, then score results with metrics.

        This is the universal entry point for all evaluation flows that involve
        running a task. The caller is responsible for resolving dataset items
        and building the execution policy.
        """
        resolved_scoring_key_mapping: ScoringKeyMappingType = (
            scoring_key_mapping if scoring_key_mapping is not None else {}
        )

        regular_metrics, task_span_metrics = (
            metrics_evaluator.split_into_regular_and_task_span_metrics(scoring_metrics)
        )

        return self._execute_evaluation(
            dataset_items=dataset_items,
            task=task,
            experiment_=experiment_,
            regular_metrics=regular_metrics,
            task_span_metrics=task_span_metrics,
            scoring_key_mapping=resolved_scoring_key_mapping,
            evaluator_model=evaluator_model,
            description=description,
            total_items=total_items,
            default_execution_policy=default_execution_policy,
            show_scores_in_progress_bar=show_scores_in_progress_bar,
        )

    def score_test_cases(
        self,
        test_cases: List[test_case.TestCase],
        scoring_metrics: List[base_metric.BaseMetric],
        scoring_key_mapping: Optional[ScoringKeyMappingType],
    ) -> List[test_result.TestResult]:
        """
        Score existing test cases with metrics (no task execution).

        This is the universal entry point for re-scoring existing experiment
        results with new metrics.
        """
        resolved_scoring_key_mapping: ScoringKeyMappingType = (
            scoring_key_mapping if scoring_key_mapping is not None else {}
        )

        regular_metrics, _ = metrics_evaluator.split_into_regular_and_task_span_metrics(
            scoring_metrics
        )

        evaluation_tasks: List[EvaluationTask[test_result.TestResult]] = [
            functools.partial(
                self._compute_test_result_for_test_case,
                test_case_=test_case_,
                regular_metrics=regular_metrics,
                scoring_key_mapping=resolved_scoring_key_mapping,
                evaluator_model=None,
            )
            for test_case_ in test_cases
        ]

        test_results: List[test_result.TestResult] = evaluation_tasks_executor.execute(
            evaluation_tasks=evaluation_tasks,
            workers=self._workers,
            verbose=self._verbose,
        )

        return test_results



import contextvars
from contextlib import contextmanager
from typing import Optional, Generator

_active_config_mask_var: contextvars.ContextVar[Optional[str]] = contextvars.ContextVar(
    "opik_active_config_mask", default=None
)


@contextmanager
def agent_config_context(mask_id: Optional[str]) -> Generator[None, None, None]:
    token = _active_config_mask_var.set(mask_id)
    try:
        yield
    finally:
        _active_config_mask_var.reset(token)


def get_active_config_mask() -> Optional[str]:
    return _active_config_mask_var.get()



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.dashboard_page_public import DashboardPagePublic
from ..types.dashboard_public import DashboardPublic
from ..types.dashboard_update_public_type import DashboardUpdatePublicType
from ..types.dashboard_write_type import DashboardWriteType
from ..types.json_node_public import JsonNodePublic
from ..types.json_node_write import JsonNodeWrite
from .raw_client import AsyncRawInsightsViewsClient, RawInsightsViewsClient

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class InsightsViewsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawInsightsViewsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawInsightsViewsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawInsightsViewsClient
        """
        return self._raw_client

    def find_insights_views(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        project_id: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DashboardPagePublic:
        """
        Find insights views in a workspace

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        project_id : typing.Optional[str]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DashboardPagePublic
            Insights view page

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.insights_views.find_insights_views()
        """
        _response = self._raw_client.find_insights_views(
            page=page,
            size=size,
            name=name,
            project_id=project_id,
            sorting=sorting,
            filters=filters,
            request_options=request_options,
        )
        return _response.data

    def create_insights_view(
        self,
        *,
        name: str,
        config: JsonNodeWrite,
        project_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        type: typing.Optional[DashboardWriteType] = OMIT,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DashboardPublic:
        """
        Create a new insights view in a workspace

        Parameters
        ----------
        name : str

        config : JsonNodeWrite

        project_id : typing.Optional[str]
            Project ID. Takes precedence over project_name when both are provided.

        project_name : typing.Optional[str]
            For project scope, specify either project_id or project_name. If project_name is provided and the project does not exist, it will be created. Ignored when project_id is provided. If neither is provided, the dashboard is created at workspace level.

        type : typing.Optional[DashboardWriteType]

        description : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DashboardPublic
            Created

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.insights_views.create_insights_view(name='name', config={'key': 'value'
        }, )
        """
        _response = self._raw_client.create_insights_view(
            name=name,
            config=config,
            project_id=project_id,
            project_name=project_name,
            type=type,
            description=description,
            request_options=request_options,
        )
        return _response.data

    def get_insights_view_by_id(
        self, insights_view_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> DashboardPublic:
        """
        Get insights view by id

        Parameters
        ----------
        insights_view_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DashboardPublic
            Insights view resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.insights_views.get_insights_view_by_id(insights_view_id='insightsViewId', )
        """
        _response = self._raw_client.get_insights_view_by_id(insights_view_id, request_options=request_options)
        return _response.data

    def delete_insights_view(
        self, insights_view_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete insights view by id

        Parameters
        ----------
        insights_view_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.insights_views.delete_insights_view(insights_view_id='insightsViewId', )
        """
        _response = self._raw_client.delete_insights_view(insights_view_id, request_options=request_options)
        return _response.data

    def update_insights_view(
        self,
        insights_view_id: str,
        *,
        name: typing.Optional[str] = OMIT,
        type: typing.Optional[DashboardUpdatePublicType] = OMIT,
        description: typing.Optional[str] = OMIT,
        config: typing.Optional[JsonNodePublic] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DashboardPublic:
        """
        Update insights view by id. Partial updates are supported - only provided fields will be updated.

        Parameters
        ----------
        insights_view_id : str

        name : typing.Optional[str]

        type : typing.Optional[DashboardUpdatePublicType]

        description : typing.Optional[str]

        config : typing.Optional[JsonNodePublic]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DashboardPublic
            Updated insights view

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.insights_views.update_insights_view(insights_view_id='insightsViewId', )
        """
        _response = self._raw_client.update_insights_view(
            insights_view_id,
            name=name,
            type=type,
            description=description,
            config=config,
            request_options=request_options,
        )
        return _response.data

    def delete_insights_views_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete insights views batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.insights_views.delete_insights_views_batch(ids=['ids'], )
        """
        _response = self._raw_client.delete_insights_views_batch(ids=ids, request_options=request_options)
        return _response.data


class AsyncInsightsViewsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawInsightsViewsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawInsightsViewsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawInsightsViewsClient
        """
        return self._raw_client

    async def find_insights_views(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        project_id: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DashboardPagePublic:
        """
        Find insights views in a workspace

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        project_id : typing.Optional[str]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DashboardPagePublic
            Insights view page

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.insights_views.find_insights_views()
        asyncio.run(main())
        """
        _response = await self._raw_client.find_insights_views(
            page=page,
            size=size,
            name=name,
            project_id=project_id,
            sorting=sorting,
            filters=filters,
            request_options=request_options,
        )
        return _response.data

    async def create_insights_view(
        self,
        *,
        name: str,
        config: JsonNodeWrite,
        project_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        type: typing.Optional[DashboardWriteType] = OMIT,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DashboardPublic:
        """
        Create a new insights view in a workspace

        Parameters
        ----------
        name : str

        config : JsonNodeWrite

        project_id : typing.Optional[str]
            Project ID. Takes precedence over project_name when both are provided.

        project_name : typing.Optional[str]
            For project scope, specify either project_id or project_name. If project_name is provided and the project does not exist, it will be created. Ignored when project_id is provided. If neither is provided, the dashboard is created at workspace level.

        type : typing.Optional[DashboardWriteType]

        description : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DashboardPublic
            Created

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.insights_views.create_insights_view(name='name', config={'key': 'value'
            }, )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_insights_view(
            name=name,
            config=config,
            project_id=project_id,
            project_name=project_name,
            type=type,
            description=description,
            request_options=request_options,
        )
        return _response.data

    async def get_insights_view_by_id(
        self, insights_view_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> DashboardPublic:
        """
        Get insights view by id

        Parameters
        ----------
        insights_view_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DashboardPublic
            Insights view resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.insights_views.get_insights_view_by_id(insights_view_id='insightsViewId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_insights_view_by_id(insights_view_id, request_options=request_options)
        return _response.data

    async def delete_insights_view(
        self, insights_view_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete insights view by id

        Parameters
        ----------
        insights_view_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.insights_views.delete_insights_view(insights_view_id='insightsViewId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_insights_view(insights_view_id, request_options=request_options)
        return _response.data

    async def update_insights_view(
        self,
        insights_view_id: str,
        *,
        name: typing.Optional[str] = OMIT,
        type: typing.Optional[DashboardUpdatePublicType] = OMIT,
        description: typing.Optional[str] = OMIT,
        config: typing.Optional[JsonNodePublic] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DashboardPublic:
        """
        Update insights view by id. Partial updates are supported - only provided fields will be updated.

        Parameters
        ----------
        insights_view_id : str

        name : typing.Optional[str]

        type : typing.Optional[DashboardUpdatePublicType]

        description : typing.Optional[str]

        config : typing.Optional[JsonNodePublic]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DashboardPublic
            Updated insights view

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.insights_views.update_insights_view(insights_view_id='insightsViewId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.update_insights_view(
            insights_view_id,
            name=name,
            type=type,
            description=description,
            config=config,
            request_options=request_options,
        )
        return _response.data

    async def delete_insights_views_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete insights views batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.insights_views.delete_insights_views_batch(ids=['ids'], )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_insights_views_batch(ids=ids, request_options=request_options)
        return _response.data



import abc
import dataclasses
from typing import Optional

from opik.message_processing import messages

from . import types as upload_types


def message_supports_upload(message: messages.BaseMessage) -> bool:
    """Helper to check if provided message supports upload."""
    return isinstance(message, messages.CreateAttachmentMessage)


@dataclasses.dataclass
class RemainingUploadData:
    uploads: int
    bytes: int
    total_size: int


class BaseFileUploadManager(abc.ABC):
    @abc.abstractmethod
    def upload(
        self,
        message: messages.BaseMessage,
        on_upload_success: Optional[upload_types.OnUploadSuccessCallback],
        on_upload_failed: Optional[upload_types.OnUploadFailureCallback],
    ) -> None:
        pass

    @abc.abstractmethod
    def remaining_data(self) -> RemainingUploadData:
        pass

    @abc.abstractmethod
    def all_done(self) -> bool:
        pass

    @abc.abstractmethod
    def flush(self, timeout: Optional[float], sleep_time: int = 5) -> bool:
        pass

    @abc.abstractmethod
    def failed_uploads(self, timeout: Optional[float]) -> int:
        pass

    @abc.abstractmethod
    def close(self) -> None:
        pass



import logging
import time
from typing import Any, Callable, Optional

from ..rest_api import client as rest_api_client
from ..rest_api.core.api_error import ApiError
from ..rate_limit import rate_limit

LOGGER = logging.getLogger(__name__)


def ensure_rest_api_call_respecting_rate_limit(
    rest_callable: Callable[[], Any],
) -> Any:
    """
    Execute a REST API call with automatic retry on rate limit (429) errors.

    This function handles HTTP 429 rate limit errors by waiting for the duration
    specified in the response headers and retrying the request. Regular retries
    for other errors are handled by the underlying rest client.

    Args:
        rest_callable: A callable that performs the REST API call.

    Returns:
        The result of the successful REST API call.

    Raises:
        ApiError: If the error is not a 429 rate limit error.
    """
    while True:
        try:
            result = rest_callable()
            return result
        except ApiError as exception:
            if exception.status_code == 429:
                if exception.headers is not None:
                    rate_limiter = rate_limit.parse_rate_limit(exception.headers)
                    if rate_limiter is not None:
                        retry_after = rate_limiter.retry_after()
                        LOGGER.info(
                            "Rate limited (HTTP 429), retrying in %s seconds",
                            retry_after,
                        )
                        time.sleep(retry_after)
                        continue

                LOGGER.info(
                    "Rate limited (HTTP 429) with no retry-after header, retrying in 1 second"
                )
                time.sleep(1)
                continue

            raise


def resolve_project_id_by_name(
    rest_client: rest_api_client.OpikApi, project_name: str
) -> str:
    """
    Resolve a project name to its project ID.

    Args:
        rest_client: The REST API client instance.
        project_name: The name of the project.

    Returns:
        The project ID.

    Raises:
        ApiError: If the project is not found or if there's an API error.
    """
    project = rest_client.projects.retrieve_project(name=project_name)
    return project.id


def resolve_project_id_by_name_optional(
    rest_client: rest_api_client.OpikApi, project_name: Optional[str]
) -> Optional[str]:
    """
    Resolve the project ID associated with the given project name if provided, otherwise return None.

    This function attempts to resolve the ID of a project by its name using a REST client. If the
    provided project name is None, the function directly returns None without performing any resolution.

    Args:
        rest_client: A REST API client instance used to make requests
            to the backend API for project resolution.
        project_name: The name of the project for which the ID needs to be resolved,
            or None if no project name is specified.

    Returns:
        The resolved project ID as a string if the project name is provided and
            the resolution succeeds, or None otherwise.
    """
    if project_name is None:
        return None
    return resolve_project_id_by_name(rest_client, project_name)



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .retention_rule_write_retention import RetentionRuleWriteRetention

__all__ = ["RetentionRuleWriteRetention"]



"""Common utilities for export functionality."""

import csv
import dataclasses
import json
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

from rich.console import Console
from rich.table import Table

import opik.dict_utils as dict_utils
from opik.api_objects.experiment.experiment_item import ExperimentItemContent

console = Console()


def matches_name_pattern(name: str, pattern: Optional[str]) -> bool:
    """Check if a name matches the given pattern using simple string matching."""
    if pattern is None:
        return True
    # Simple string matching - check if pattern is contained in name (case-insensitive)
    return pattern.lower() in name.lower()


def serialize_experiment_item(item: ExperimentItemContent) -> Dict[str, Any]:
    """Serialize an ExperimentItemContent dataclass to a dictionary."""
    return dataclasses.asdict(item)


def should_skip_file(file_path: Path, force: bool) -> bool:
    """Check if a file should be skipped based on existence and force flag."""
    return file_path.exists() and not force


def write_csv_data(
    data: Dict[str, Any],
    file_path: Path,
    csv_row_converter_func: Callable[[Dict[str, Any]], List[Dict]],
) -> None:
    """Write data to CSV file using the provided row converter function."""
    try:
        csv_rows = csv_row_converter_func(data)
        if not csv_rows:
            raise ValueError(f"CSV row converter returned empty list for {file_path}")
        with open(file_path, "w", newline="", encoding="utf-8") as csv_file_handle:
            csv_fieldnames = list(csv_rows[0].keys())
            csv_writer = csv.DictWriter(csv_file_handle, fieldnames=csv_fieldnames)
            csv_writer.writeheader()
            csv_writer.writerows(csv_rows)
    except Exception as e:
        # Re-raise with more context
        raise RuntimeError(f"Failed to write CSV file {file_path}: {e}") from e


def write_json_data(data: Dict[str, Any], file_path: Path) -> None:
    """Write data to JSON file."""
    with open(file_path, "w", encoding="utf-8") as f:
        json.dump(data, f, indent=2, default=str)


def debug_print(message: str, debug: bool) -> None:
    """Print debug message only if debug is enabled."""
    if debug:
        console.print(f"[blue]{message}[/blue]")


def create_experiment_data_structure(
    experiment: Any, experiment_items: List[ExperimentItemContent]
) -> Dict[str, Any]:
    """Create a comprehensive experiment data structure for export."""
    # Get the full experiment data which contains all fields
    experiment_data_obj = experiment.get_experiment_data()

    return {
        "experiment": {
            "id": experiment.id,
            "name": experiment.name,
            "dataset_name": experiment.dataset_name,
            "metadata": getattr(experiment_data_obj, "metadata", None),
            "type": getattr(experiment_data_obj, "type", None),
            "status": getattr(experiment_data_obj, "status", None),
            "created_at": getattr(experiment_data_obj, "created_at", None),
            "last_updated_at": getattr(experiment_data_obj, "last_updated_at", None),
            "created_by": getattr(experiment_data_obj, "created_by", None),
            "last_updated_by": getattr(experiment_data_obj, "last_updated_by", None),
            "trace_count": getattr(experiment_data_obj, "trace_count", None),
            "total_estimated_cost": getattr(
                experiment_data_obj, "total_estimated_cost", None
            ),
            "total_estimated_cost_avg": getattr(
                experiment_data_obj, "total_estimated_cost_avg", None
            ),
            "usage": getattr(experiment_data_obj, "usage", None),
            "feedback_scores": getattr(experiment_data_obj, "feedback_scores", None),
            "comments": getattr(experiment_data_obj, "comments", None),
            "duration": getattr(experiment_data_obj, "duration", None),
            "prompt_version": getattr(experiment_data_obj, "prompt_version", None),
            "prompt_versions": getattr(experiment_data_obj, "prompt_versions", None),
        },
        "items": [serialize_experiment_item(item) for item in experiment_items],
        "downloaded_at": datetime.now().isoformat(),
    }


def dump_to_file(
    data: dict,
    file_path: Path,
    file_format: str,
    csv_writer: Optional[csv.DictWriter] = None,
    csv_fieldnames: Optional[List[str]] = None,
    data_type: str = "trace",
) -> tuple:
    """
    Helper function to dump data to file in the specified format.

    Args:
        data: The data to dump
        file_path: Path where to save the file
        file_format: Format to use ("json" or "csv")
        csv_writer: Existing CSV writer (for CSV format)
        csv_fieldnames: Existing CSV fieldnames (for CSV format)
        data_type: Type of data ("trace", "dataset", "prompt", "experiment")

    Returns:
        Tuple of (csv_writer, csv_fieldnames) for CSV format, or (None, None) for JSON
    """
    if file_format.lower() == "csv":
        # Convert to CSV rows based on data type
        if data_type == "trace":
            csv_rows = trace_to_csv_rows(data)
        elif data_type == "dataset":
            csv_rows = dataset_to_csv_rows(data)
        elif data_type == "prompt":
            csv_rows = prompt_to_csv_rows(data)
        elif data_type == "experiment":
            csv_rows = experiment_to_csv_rows(data)
        else:
            # Fallback to trace format for unknown types
            csv_rows = trace_to_csv_rows(data)

        # Initialize CSV writer if not already done
        if csv_writer is None and csv_rows:
            # Use context manager to ensure file is properly closed
            with open(file_path, "w", newline="", encoding="utf-8") as csv_file_handle:
                csv_fieldnames = list(csv_rows[0].keys())
                csv_writer = csv.DictWriter(csv_file_handle, fieldnames=csv_fieldnames)
                csv_writer.writeheader()
                # Write rows while file is open
                csv_writer.writerows(csv_rows)

            # File is closed, return None, None
            return None, None

        # Write rows to existing writer (caller manages file lifecycle)
        if csv_writer and csv_rows:
            csv_writer.writerows(csv_rows)

        return csv_writer, csv_fieldnames
    else:
        # Save to JSON file
        with open(file_path, "w", encoding="utf-8") as f:
            json.dump(data, f, indent=2, default=str)

        return None, None


def trace_to_csv_rows(trace_data: dict) -> List[Dict]:
    """Convert trace data to CSV rows format."""
    trace = trace_data["trace"]
    spans = trace_data.get("spans", [])

    # Flatten trace data with "trace" prefix
    trace_flat = dict_utils.flatten_dict(trace, parent_key="trace", delim="_")

    # If no spans, create a single row for the trace
    if not spans:
        # Create empty span fields to maintain consistent structure
        span_flat = {f"span_{key}": "" for key in trace.keys()}
        span_flat["span_parent_span_id"] = ""  # Special case for parent_span_id

        # Combine trace and empty span data
        row = {**trace_flat, **span_flat}
        return [row]

    # Create rows for each span
    rows = []
    for span in spans:
        # Flatten span data with "span" prefix
        span_flat = dict_utils.flatten_dict(span, parent_key="span", delim="_")

        # Combine trace and span data
        row = {**trace_flat, **span_flat}
        rows.append(row)

    return rows


def dataset_to_csv_rows(dataset_data: dict) -> List[Dict]:
    """Convert dataset data to CSV rows format."""
    rows = []

    # Create a row for each dataset item
    items = dataset_data.get("items", [])
    for i, item in enumerate(items):
        # Flatten item data - use all fields from the item
        # (datasets can have any user-defined keys/values)
        item_flat = dict_utils.flatten_dict(
            item,  # Use the entire item dict, not just hardcoded fields
            parent_key="item",
            delim="_",
        )

        # Create row with item data and index
        row = {**item_flat}
        row["item_index"] = i  # Add index for ordering
        rows.append(row)

    return rows


def prompt_to_csv_rows(prompt_data: dict) -> List[Dict]:
    """Convert prompt data to CSV rows format."""
    # Flatten prompt data
    prompt_flat = dict_utils.flatten_dict(prompt_data, parent_key="prompt", delim="_")

    # Create a single row for the prompt
    return [prompt_flat]


def experiment_to_csv_rows(experiment_data: dict) -> List[Dict]:
    """Convert experiment data to CSV rows format."""
    rows = []

    # Flatten experiment metadata
    experiment_flat = dict_utils.flatten_dict(
        {
            "id": experiment_data.get("experiment", {}).get("id"),
            "name": experiment_data.get("experiment", {}).get("name"),
            "dataset_name": experiment_data.get("experiment", {}).get("dataset_name"),
            "type": experiment_data.get("experiment", {}).get("type"),
            "status": experiment_data.get("experiment", {}).get("status"),
            "created_at": experiment_data.get("experiment", {}).get("created_at"),
            "last_updated_at": experiment_data.get("experiment", {}).get(
                "last_updated_at"
            ),
            "created_by": experiment_data.get("experiment", {}).get("created_by"),
            "last_updated_by": experiment_data.get("experiment", {}).get(
                "last_updated_by"
            ),
            "trace_count": experiment_data.get("experiment", {}).get("trace_count"),
            "total_estimated_cost": experiment_data.get("experiment", {}).get(
                "total_estimated_cost"
            ),
            "downloaded_at": experiment_data.get("downloaded_at"),
        },
        parent_key="experiment",
        delim="_",
    )

    # Create a row for each experiment item
    items = experiment_data.get("items", [])
    for i, item in enumerate(items):
        # Flatten item data
        item_flat = dict_utils.flatten_dict(
            {
                "id": item.get("id"),
                "experiment_id": item.get("experiment_id"),
                "dataset_item_id": item.get("dataset_item_id"),
                "trace_id": item.get("trace_id"),
                "input": item.get("input"),
                "output": item.get("output"),
                "feedback_scores": item.get("feedback_scores"),
                "comments": item.get("comments"),
                "total_estimated_cost": item.get("total_estimated_cost"),
                "duration": item.get("duration"),
                "usage": item.get("usage"),
                "created_at": item.get("created_at"),
                "last_updated_at": item.get("last_updated_at"),
                "created_by": item.get("created_by"),
                "last_updated_by": item.get("last_updated_by"),
                "trace_visibility_mode": item.get("trace_visibility_mode"),
            },
            parent_key="item",
            delim="_",
        )

        # Combine experiment and item data
        row = {**experiment_flat, **item_flat}
        row["item_index"] = i  # Add index for ordering
        rows.append(row)

    # If no items, return just the experiment metadata
    if not items:
        rows.append(experiment_flat)

    return rows


def print_export_summary(stats: Dict[str, int], format: str = "json") -> None:
    """Print a nice summary table of export statistics."""
    table = Table(
        title="📊 Export Summary", show_header=True, header_style="bold magenta"
    )
    table.add_column("Type", style="cyan", no_wrap=True)
    table.add_column("Exported", justify="right", style="green")
    table.add_column("Skipped", justify="right", style="yellow")
    table.add_column("Files", style="blue")

    # Add rows for each type
    if stats.get("experiments", 0) > 0 or stats.get("experiments_skipped", 0) > 0:
        exported = stats.get("experiments", 0)
        skipped = stats.get("experiments_skipped", 0)
        experiment_file_pattern = (
            "experiments_*.csv" if format.lower() == "csv" else "experiment_*.json"
        )
        table.add_row(
            "🧪 Experiments",
            str(exported),
            str(skipped) if skipped > 0 else "",
            experiment_file_pattern,
        )

    if stats.get("datasets", 0) > 0 or stats.get("datasets_skipped", 0) > 0:
        exported = stats.get("datasets", 0)
        skipped = stats.get("datasets_skipped", 0)
        dataset_file_pattern = (
            "dataset_*.csv" if format.lower() == "csv" else "dataset_*.json"
        )
        table.add_row(
            "📊 Datasets",
            str(exported),
            str(skipped) if skipped > 0 else "",
            dataset_file_pattern,
        )

    if stats.get("traces", 0) > 0 or stats.get("traces_skipped", 0) > 0:
        exported = stats.get("traces", 0)
        skipped = stats.get("traces_skipped", 0)
        trace_file_pattern = (
            "trace_*.csv" if format.lower() == "csv" else "trace_*.json"
        )
        table.add_row(
            "🔍 Traces",
            str(exported),
            str(skipped) if skipped > 0 else "",
            trace_file_pattern,
        )

    if stats.get("prompts", 0) > 0 or stats.get("prompts_skipped", 0) > 0:
        exported = stats.get("prompts", 0)
        skipped = stats.get("prompts_skipped", 0)
        prompt_file_pattern = (
            "prompts_*.csv" if format.lower() == "csv" else "prompt_*.json"
        )
        table.add_row(
            "💬 Prompts",
            str(exported),
            str(skipped) if skipped > 0 else "",
            prompt_file_pattern,
        )

    if stats.get("projects", 0) > 0 or stats.get("projects_skipped", 0) > 0:
        exported = stats.get("projects", 0)
        skipped = stats.get("projects_skipped", 0)
        table.add_row(
            "📁 Projects",
            str(exported),
            str(skipped) if skipped > 0 else "",
            "project directories",
        )

    # Calculate totals
    total_exported = sum(
        [
            stats.get(key, 0)
            for key in ["experiments", "datasets", "traces", "prompts", "projects"]
        ]
    )
    total_skipped = sum(
        [
            stats.get(key, 0)
            for key in [
                "experiments_skipped",
                "datasets_skipped",
                "traces_skipped",
                "prompts_skipped",
                "projects_skipped",
            ]
        ]
    )
    total_files = total_exported + total_skipped

    table.add_row("", "", "", "", style="bold")
    table.add_row(
        "📦 Total",
        str(total_exported),
        str(total_skipped) if total_skipped > 0 else "",
        f"{total_files} files",
        style="bold green",
    )

    console.print()
    console.print(table)
    console.print()



import logging
import threading
import time
from typing import Optional

from opik.healthcheck import connection_monitor
from . import db_manager, types
from .. import messages


LOGGER = logging.getLogger(__name__)


def _check_message_id(message_id: Optional[int]) -> None:
    if message_id is None:
        # fail fast - it is a programming error if message ID is not provided
        raise ValueError("Message ID expected")


class ReplayManager(threading.Thread):
    """
    Manages replaying messages to the server for a connection management system.

    The ReplayManager is responsible for ensuring that messages which fail to be
    delivered due to dropped connections are replayed to the server
    when the connection is restored. It continuously monitors the connection status
    and triggers failed message replays as necessary. The class runs in its own
    thread, leveraging the threading.Thread base class.
    """

    def __init__(
        self,
        monitor: connection_monitor.OpikConnectionMonitor,
        batch_size: int,
        batch_replay_delay: float,
        tick_interval_seconds: float,
    ):
        """
        Initializes the ReplayManager instance.

        Creates the replay manager thread with a specified connection monitor and tick interval.

        Args:
            monitor: An instance of OpikConnectionMonitor used for monitoring the connection.
            batch_size: The size of batches for processing.
            batch_replay_delay: The delay (in seconds) between replaying batches of messages.
            tick_interval_seconds: Interval in seconds between execution ticks. Default is 0.3.
        """
        super().__init__(daemon=True, name="ReplayManager")
        self._stop_running = threading.Event()
        self._monitor = monitor
        self._replay_callback: Optional[types.ReplayCallback] = None
        self._tick_interval_seconds = tick_interval_seconds
        self._next_tick_time = time.time() + self._tick_interval_seconds

        self._next_message_id = 0
        self._message_id_lock = threading.RLock()

        self._db_manager = db_manager.DBManager(
            batch_size=batch_size,
            batch_replay_delay=batch_replay_delay,
        )

    def start(self) -> None:
        self._check_replay_callback()
        super().start()

    def run(self) -> None:
        try:
            while not self._stop_running.is_set():
                self._loop()
        finally:
            # release the database connection
            self._db_manager.close()

    @property
    def database_manager(self) -> db_manager.DBManager:
        return self._db_manager

    @property
    def has_server_connection(self) -> bool:
        """Checks if SDK has a connection to the OPIK server."""
        return self._monitor.has_server_connection

    def register_message(
        self,
        message: messages.BaseMessage,
        status: db_manager.MessageStatus = db_manager.MessageStatus.registered,
    ) -> None:
        """Registers a message to be replayed if the connection is lost."""
        with self._message_id_lock:
            # set message ID if not set yet
            if message.message_id is None:
                message.message_id = self._next_message_id
                self._next_message_id += 1

        try:
            self._db_manager.register_message(message, status=status)
        except Exception as ex:
            LOGGER.error(
                "Failed to register message for replay, reason: %s", ex, exc_info=True
            )

    def unregister_message(self, message_id: int) -> None:
        """Unregisters a message from being replayed if the connection is lost."""
        _check_message_id(message_id)
        try:
            self._db_manager.update_message(
                message_id, status=db_manager.MessageStatus.delivered
            )
        except Exception as ex:
            LOGGER.error(
                "Failed to un-register message from replay, reason: %s",
                ex,
                exc_info=True,
            )

    def message_sent_failed(
        self, message_id: int, failure_reason: Optional[str] = None
    ) -> None:
        """Notifies the manager that a message was not sent due to a connection failure."""
        _check_message_id(message_id)
        try:
            self._monitor.connection_failed(failure_reason=failure_reason)
            self._db_manager.update_message(
                message_id, status=db_manager.MessageStatus.failed
            )
        except Exception as ex:
            LOGGER.error(
                "Failed to mark message as send-failed in replay manager, reason: %s",
                ex,
                exc_info=True,
            )

    def set_replay_callback(self, callback: types.ReplayCallback) -> None:
        """Sets the callback to be invoked when replaying failed messages."""
        self._replay_callback = callback

    def close(self) -> None:
        """Stop the replay manager."""
        self._stop_running.set()

    def flush(self) -> None:
        """Force replay of all failed messages to the server."""
        self._check_replay_callback()
        # ignore MyPy check because already asserted above
        self._replay_failed_messages()

    def _loop(self) -> None:
        sleep_time = self._next_tick_time - time.time()
        if sleep_time > 0:
            # sleep until the next tick time to avoid excessive CPU usage - interruptible-sleep by close
            self._stop_running.wait(sleep_time)

        try:
            status = self._monitor.tick()
            if status == connection_monitor.ConnectionStatus.connection_restored:
                # the connection was restored, replay all failed messages
                self._replay_failed_messages()
                self._monitor.reset()
        except Exception as ex:
            LOGGER.warning(
                "Failed to tick the connection monitor or replay failed messages, will repeat. Reason: %s",
                ex,
                exc_info=ex,
            )

        # schedule the next tick after the potential replay above
        self._next_tick_time = time.time() + self._tick_interval_seconds

    def _check_replay_callback(self) -> None:
        if self._replay_callback is None:
            raise ValueError(
                "Replay callback must be set before starting the replay manager"
            )

    def _replay_failed_messages(self) -> None:
        self._check_replay_callback()
        try:
            # ignore MyPy check because already asserted above
            replayed = self._db_manager.replay_failed_messages(
                self._replay_callback  # type: ignore
            )

            if replayed > 0:
                LOGGER.info(
                    "Replayed %d message(s) that were not sent due to server connection issues",
                    replayed,
                )
        except Exception as ex:
            LOGGER.error(
                "Failed to replay failed messages, reason: %s", ex, exc_info=True
            )



from .metric import LLMJudge
from .config import (
    LLMJudgeConfig,
    LLMJudgeModelConfig,
    LLMJudgeMessage,
    LLMJudgeSchemaItem,
    DEFAULT_MODEL_NAME,
)
from .parsers import (
    AssertionResultItem,
    ResponseSchema,
)

__all__ = [
    "LLMJudge",
    "LLMJudgeConfig",
    "LLMJudgeModelConfig",
    "LLMJudgeMessage",
    "LLMJudgeSchemaItem",
    "DEFAULT_MODEL_NAME",
    "AssertionResultItem",
    "ResponseSchema",
]



"""Rule-based tone metric for assistant responses."""

from __future__ import annotations

import re
from typing import Any, Iterable, Optional, Sequence

from opik.exceptions import MetricComputationError
from opik.evaluation.metrics.base_metric import BaseMetric
from opik.evaluation.metrics.score_result import ScoreResult

# Default tone lexicons/phrases kept inline for easier discoverability.
_POSITIVE_LEXICON = {
    "appreciate",
    "assist",
    "glad",
    "helpful",
    "please",
    "thank",
    "welcome",
    "happy",
    "support",
    "great",
    "excellent",
    "wonderful",
}

_NEGATIVE_LEXICON = {
    "angry",
    "awful",
    "bad",
    "complain",
    "frustrated",
    "hate",
    "incompetent",
    "terrible",
    "useless",
    "stupid",
    "idiot",
}

_FORBIDDEN_PHRASES = {
    "shut up",
    "this is pointless",
    "not my problem",
    "i refuse to assist",
}


class Tone(BaseMetric):
    """
    Flag tone issues like excessive negativity, shouting, or forbidden phrases.

    Args:
        name: Display name for the metric result. Defaults to ``"tone_metric"``.
        track: Whether to automatically track results. Defaults to ``True``.
        project_name: Optional tracking project name. Defaults to ``None``.
        min_sentiment: Minimum sentiment score required (``-1.0`` to ``1.0`` scale).
        max_upper_ratio: Maximum allowed ratio of uppercase characters.
        max_exclamations: Cap on the number of exclamation marks.
        positive_lexicon: Optional iterable of positive tokens counted for sentiment.
        negative_lexicon: Optional iterable of negative tokens counted for sentiment.
        forbidden_phrases: Optional sequence of phrases that immediately fail the
            check.

    Example:
        >>> from opik.evaluation.metrics import Tone
        >>> metric = Tone(max_exclamations=2)
        >>> result = metric.score("THANK YOU for your patience!!!")
        >>> result.value  # doctest: +SKIP
        0.0
    """

    def __init__(
        self,
        name: str = "tone_metric",
        track: bool = True,
        project_name: Optional[str] = None,
        min_sentiment: float = -0.2,
        max_upper_ratio: float = 0.3,
        max_exclamations: int = 3,
        positive_lexicon: Optional[Iterable[str]] = None,
        negative_lexicon: Optional[Iterable[str]] = None,
        forbidden_phrases: Optional[Sequence[str]] = None,
    ) -> None:
        super().__init__(name=name, track=track, project_name=project_name)
        self._min_sentiment = min_sentiment
        self._max_upper_ratio = max_upper_ratio
        self._max_exclamations = max_exclamations
        self._positive = set(
            word.lower() for word in (positive_lexicon or _POSITIVE_LEXICON)
        )
        self._negative = set(
            word.lower() for word in (negative_lexicon or _NEGATIVE_LEXICON)
        )
        phrases = forbidden_phrases or _FORBIDDEN_PHRASES
        self._forbidden = [phrase.lower() for phrase in phrases]

    def score(self, output: str, **ignored_kwargs: Any) -> ScoreResult:
        if not output or not output.strip():
            raise MetricComputationError("Text is empty (Tone metric).")

        tokens = re.findall(r"\b\w+\b", output.lower())
        if not tokens:
            raise MetricComputationError("Unable to tokenize text for Tone metric.")

        sentiment_score = self._compute_sentiment(tokens)
        upper_ratio = _uppercase_ratio(output)
        exclamation_count = output.count("!")
        forbidden_hit = any(phrase in output.lower() for phrase in self._forbidden)

        passes = (
            sentiment_score >= self._min_sentiment
            and upper_ratio <= self._max_upper_ratio
            and exclamation_count <= self._max_exclamations
            and not forbidden_hit
        )

        metadata = {
            "sentiment_score": sentiment_score,
            "uppercase_ratio": upper_ratio,
            "exclamation_count": exclamation_count,
            "forbidden_hit": forbidden_hit,
            "thresholds": {
                "min_sentiment": self._min_sentiment,
                "max_upper_ratio": self._max_upper_ratio,
                "max_exclamations": self._max_exclamations,
            },
        }

        reason = (
            "Tone is within configured guardrails"
            if passes
            else "Tone violates guardrails"
        )
        value = 1.0 if passes else 0.0
        return ScoreResult(
            value=value, name=self.name, reason=reason, metadata=metadata
        )

    def _compute_sentiment(self, tokens: Sequence[str]) -> float:
        pos_hits = sum(token in self._positive for token in tokens)
        neg_hits = sum(token in self._negative for token in tokens)
        total = pos_hits + neg_hits
        if total == 0:
            return 0.0
        return (pos_hits - neg_hits) / total


def _uppercase_ratio(text: str) -> float:
    letters = [char for char in text if char.isalpha()]
    if not letters:
        return 0.0
    upper = sum(1 for char in letters if char.isupper())
    return upper / len(letters)



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .feedback_score_average import FeedbackScoreAverage
from .percentage_values import PercentageValues


class AggregationData(UniversalBaseModel):
    experiment_count: typing.Optional[int] = None
    trace_count: typing.Optional[int] = None
    total_estimated_cost: typing.Optional[float] = None
    total_estimated_cost_avg: typing.Optional[float] = None
    duration: typing.Optional[PercentageValues] = None
    feedback_scores: typing.Optional[typing.List[FeedbackScoreAverage]] = None
    experiment_scores: typing.Optional[typing.List[FeedbackScoreAverage]] = None
    pass_rate_avg: typing.Optional[float] = None
    passed_count_sum: typing.Optional[int] = None
    total_count_sum: typing.Optional[int] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import datetime
from typing import Optional


def local_timestamp() -> datetime.datetime:
    now = datetime.datetime.now(datetime.timezone.utc)
    return now


def datetime_to_iso8601(value: datetime.datetime) -> str:
    return value.isoformat()


def parse_iso_timestamp(timestamp_str: Optional[str]) -> Optional[datetime.datetime]:
    """Parse an ISO 8601 timestamp string to datetime."""
    if timestamp_str is None:
        return None
    try:
        timestamp_str = timestamp_str.replace("Z", "+00:00")
        return datetime.datetime.fromisoformat(timestamp_str)
    except (ValueError, TypeError):
        return None



# This file was auto-generated by Fern from our API Definition.

import contextlib
import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..core.serialization import convert_and_respect_annotation_metadata
from ..errors.forbidden_error import ForbiddenError
from ..errors.unauthorized_error import UnauthorizedError
from ..types.attachment_page import AttachmentPage
from ..types.complete_multipart_upload_request_entity_type import CompleteMultipartUploadRequestEntityType
from ..types.error_message import ErrorMessage
from ..types.multipart_upload_part import MultipartUploadPart
from ..types.start_multipart_upload_response import StartMultipartUploadResponse
from .types.attachment_list_request_entity_type import AttachmentListRequestEntityType
from .types.download_attachment_request_entity_type import DownloadAttachmentRequestEntityType
from .types.start_multipart_upload_request_entity_type import StartMultipartUploadRequestEntityType
from .types.upload_attachment_request_entity_type import UploadAttachmentRequestEntityType

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawAttachmentsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def attachment_list(
        self,
        *,
        project_id: str,
        entity_type: AttachmentListRequestEntityType,
        entity_id: str,
        path: str,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[AttachmentPage]:
        """
        Attachments list for entity

        Parameters
        ----------
        project_id : str

        entity_type : AttachmentListRequestEntityType

        entity_id : str

        path : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AttachmentPage]
            Attachment Resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/attachment/list",
            method="GET",
            params={
                "page": page,
                "size": size,
                "project_id": project_id,
                "entity_type": entity_type,
                "entity_id": entity_id,
                "path": path,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AttachmentPage,
                    parse_obj_as(
                        type_=AttachmentPage,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def complete_multi_part_upload(
        self,
        *,
        file_name: str,
        entity_type: CompleteMultipartUploadRequestEntityType,
        entity_id: str,
        file_size: int,
        upload_id: str,
        uploaded_file_parts: typing.Sequence[MultipartUploadPart],
        project_name: typing.Optional[str] = OMIT,
        container_id: typing.Optional[str] = OMIT,
        mime_type: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Complete multipart attachment upload

        Parameters
        ----------
        file_name : str

        entity_type : CompleteMultipartUploadRequestEntityType

        entity_id : str

        file_size : int

        upload_id : str

        uploaded_file_parts : typing.Sequence[MultipartUploadPart]

        project_name : typing.Optional[str]
            If null, the default project is used

        container_id : typing.Optional[str]

        mime_type : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/attachment/upload-complete",
            method="POST",
            json={
                "file_name": file_name,
                "project_name": project_name,
                "entity_type": entity_type,
                "entity_id": entity_id,
                "container_id": container_id,
                "file_size": file_size,
                "mime_type": mime_type,
                "upload_id": upload_id,
                "uploaded_file_parts": convert_and_respect_annotation_metadata(
                    object_=uploaded_file_parts, annotation=typing.Sequence[MultipartUploadPart], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete_attachments(
        self,
        *,
        file_name: str,
        entity_type: CompleteMultipartUploadRequestEntityType,
        entity_id: str,
        file_size: int,
        upload_id: str,
        uploaded_file_parts: typing.Sequence[MultipartUploadPart],
        project_name: typing.Optional[str] = OMIT,
        container_id: typing.Optional[str] = OMIT,
        mime_type: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Delete attachments

        Parameters
        ----------
        file_name : str

        entity_type : CompleteMultipartUploadRequestEntityType

        entity_id : str

        file_size : int

        upload_id : str

        uploaded_file_parts : typing.Sequence[MultipartUploadPart]

        project_name : typing.Optional[str]
            If null, the default project is used

        container_id : typing.Optional[str]

        mime_type : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/attachment/delete",
            method="POST",
            json={
                "file_name": file_name,
                "project_name": project_name,
                "entity_type": entity_type,
                "entity_id": entity_id,
                "container_id": container_id,
                "file_size": file_size,
                "mime_type": mime_type,
                "upload_id": upload_id,
                "uploaded_file_parts": convert_and_respect_annotation_metadata(
                    object_=uploaded_file_parts, annotation=typing.Sequence[MultipartUploadPart], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    @contextlib.contextmanager
    def download_attachment(
        self,
        *,
        container_id: str,
        entity_type: DownloadAttachmentRequestEntityType,
        entity_id: str,
        file_name: str,
        mime_type: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.Iterator[HttpResponse[typing.Iterator[bytes]]]:
        """
        Download attachment from MinIO

        Parameters
        ----------
        container_id : str

        entity_type : DownloadAttachmentRequestEntityType

        entity_id : str

        file_name : str

        mime_type : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.Iterator[HttpResponse[typing.Iterator[bytes]]]
            Attachment Resource
        """
        with self._client_wrapper.httpx_client.stream(
            "v1/private/attachment/download",
            method="GET",
            params={
                "workspace_name": workspace_name,
                "container_id": container_id,
                "entity_type": entity_type,
                "entity_id": entity_id,
                "file_name": file_name,
                "mime_type": mime_type,
            },
            request_options=request_options,
        ) as _response:

            def stream() -> HttpResponse[typing.Iterator[bytes]]:
                try:
                    if 200 <= _response.status_code < 300:
                        _chunk_size = request_options.get("chunk_size", None) if request_options is not None else None
                        return HttpResponse(
                            response=_response, data=(_chunk for _chunk in _response.iter_bytes(chunk_size=_chunk_size))
                        )
                    _response.read()
                    if _response.status_code == 401:
                        raise UnauthorizedError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                typing.Optional[typing.Any],
                                parse_obj_as(
                                    type_=typing.Optional[typing.Any],  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    if _response.status_code == 403:
                        raise ForbiddenError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                ErrorMessage,
                                parse_obj_as(
                                    type_=ErrorMessage,  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    _response_json = _response.json()
                except JSONDecodeError:
                    raise ApiError(
                        status_code=_response.status_code, headers=dict(_response.headers), body=_response.text
                    )
                raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

            yield stream()

    def start_multi_part_upload(
        self,
        *,
        file_name: str,
        num_of_file_parts: int,
        entity_type: StartMultipartUploadRequestEntityType,
        entity_id: str,
        path: str,
        mime_type: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[StartMultipartUploadResponse]:
        """
        Start multipart attachment upload

        Parameters
        ----------
        file_name : str

        num_of_file_parts : int

        entity_type : StartMultipartUploadRequestEntityType

        entity_id : str

        path : str

        mime_type : typing.Optional[str]

        project_name : typing.Optional[str]
            If null, the default project is used

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[StartMultipartUploadResponse]
            MultipartUploadResponse
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/attachment/upload-start",
            method="POST",
            json={
                "file_name": file_name,
                "num_of_file_parts": num_of_file_parts,
                "mime_type": mime_type,
                "project_name": project_name,
                "entity_type": entity_type,
                "entity_id": entity_id,
                "path": path,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    StartMultipartUploadResponse,
                    parse_obj_as(
                        type_=StartMultipartUploadResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def upload_attachment(
        self,
        *,
        file_name: str,
        entity_type: UploadAttachmentRequestEntityType,
        entity_id: str,
        request: typing.Dict[str, typing.Optional[typing.Any]],
        project_name: typing.Optional[str] = None,
        mime_type: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Upload attachment to MinIO

        Parameters
        ----------
        file_name : str

        entity_type : UploadAttachmentRequestEntityType

        entity_id : str

        request : typing.Dict[str, typing.Optional[typing.Any]]

        project_name : typing.Optional[str]

        mime_type : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/attachment/upload",
            method="PUT",
            params={
                "file_name": file_name,
                "project_name": project_name,
                "mime_type": mime_type,
                "entity_type": entity_type,
                "entity_id": entity_id,
            },
            json=request,
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawAttachmentsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def attachment_list(
        self,
        *,
        project_id: str,
        entity_type: AttachmentListRequestEntityType,
        entity_id: str,
        path: str,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[AttachmentPage]:
        """
        Attachments list for entity

        Parameters
        ----------
        project_id : str

        entity_type : AttachmentListRequestEntityType

        entity_id : str

        path : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AttachmentPage]
            Attachment Resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/attachment/list",
            method="GET",
            params={
                "page": page,
                "size": size,
                "project_id": project_id,
                "entity_type": entity_type,
                "entity_id": entity_id,
                "path": path,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AttachmentPage,
                    parse_obj_as(
                        type_=AttachmentPage,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def complete_multi_part_upload(
        self,
        *,
        file_name: str,
        entity_type: CompleteMultipartUploadRequestEntityType,
        entity_id: str,
        file_size: int,
        upload_id: str,
        uploaded_file_parts: typing.Sequence[MultipartUploadPart],
        project_name: typing.Optional[str] = OMIT,
        container_id: typing.Optional[str] = OMIT,
        mime_type: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Complete multipart attachment upload

        Parameters
        ----------
        file_name : str

        entity_type : CompleteMultipartUploadRequestEntityType

        entity_id : str

        file_size : int

        upload_id : str

        uploaded_file_parts : typing.Sequence[MultipartUploadPart]

        project_name : typing.Optional[str]
            If null, the default project is used

        container_id : typing.Optional[str]

        mime_type : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/attachment/upload-complete",
            method="POST",
            json={
                "file_name": file_name,
                "project_name": project_name,
                "entity_type": entity_type,
                "entity_id": entity_id,
                "container_id": container_id,
                "file_size": file_size,
                "mime_type": mime_type,
                "upload_id": upload_id,
                "uploaded_file_parts": convert_and_respect_annotation_metadata(
                    object_=uploaded_file_parts, annotation=typing.Sequence[MultipartUploadPart], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete_attachments(
        self,
        *,
        file_name: str,
        entity_type: CompleteMultipartUploadRequestEntityType,
        entity_id: str,
        file_size: int,
        upload_id: str,
        uploaded_file_parts: typing.Sequence[MultipartUploadPart],
        project_name: typing.Optional[str] = OMIT,
        container_id: typing.Optional[str] = OMIT,
        mime_type: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Delete attachments

        Parameters
        ----------
        file_name : str

        entity_type : CompleteMultipartUploadRequestEntityType

        entity_id : str

        file_size : int

        upload_id : str

        uploaded_file_parts : typing.Sequence[MultipartUploadPart]

        project_name : typing.Optional[str]
            If null, the default project is used

        container_id : typing.Optional[str]

        mime_type : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/attachment/delete",
            method="POST",
            json={
                "file_name": file_name,
                "project_name": project_name,
                "entity_type": entity_type,
                "entity_id": entity_id,
                "container_id": container_id,
                "file_size": file_size,
                "mime_type": mime_type,
                "upload_id": upload_id,
                "uploaded_file_parts": convert_and_respect_annotation_metadata(
                    object_=uploaded_file_parts, annotation=typing.Sequence[MultipartUploadPart], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    @contextlib.asynccontextmanager
    async def download_attachment(
        self,
        *,
        container_id: str,
        entity_type: DownloadAttachmentRequestEntityType,
        entity_id: str,
        file_name: str,
        mime_type: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.AsyncIterator[AsyncHttpResponse[typing.AsyncIterator[bytes]]]:
        """
        Download attachment from MinIO

        Parameters
        ----------
        container_id : str

        entity_type : DownloadAttachmentRequestEntityType

        entity_id : str

        file_name : str

        mime_type : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.AsyncIterator[AsyncHttpResponse[typing.AsyncIterator[bytes]]]
            Attachment Resource
        """
        async with self._client_wrapper.httpx_client.stream(
            "v1/private/attachment/download",
            method="GET",
            params={
                "workspace_name": workspace_name,
                "container_id": container_id,
                "entity_type": entity_type,
                "entity_id": entity_id,
                "file_name": file_name,
                "mime_type": mime_type,
            },
            request_options=request_options,
        ) as _response:

            async def stream() -> AsyncHttpResponse[typing.AsyncIterator[bytes]]:
                try:
                    if 200 <= _response.status_code < 300:
                        _chunk_size = request_options.get("chunk_size", None) if request_options is not None else None
                        return AsyncHttpResponse(
                            response=_response,
                            data=(_chunk async for _chunk in _response.aiter_bytes(chunk_size=_chunk_size)),
                        )
                    await _response.aread()
                    if _response.status_code == 401:
                        raise UnauthorizedError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                typing.Optional[typing.Any],
                                parse_obj_as(
                                    type_=typing.Optional[typing.Any],  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    if _response.status_code == 403:
                        raise ForbiddenError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                ErrorMessage,
                                parse_obj_as(
                                    type_=ErrorMessage,  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    _response_json = _response.json()
                except JSONDecodeError:
                    raise ApiError(
                        status_code=_response.status_code, headers=dict(_response.headers), body=_response.text
                    )
                raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

            yield await stream()

    async def start_multi_part_upload(
        self,
        *,
        file_name: str,
        num_of_file_parts: int,
        entity_type: StartMultipartUploadRequestEntityType,
        entity_id: str,
        path: str,
        mime_type: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[StartMultipartUploadResponse]:
        """
        Start multipart attachment upload

        Parameters
        ----------
        file_name : str

        num_of_file_parts : int

        entity_type : StartMultipartUploadRequestEntityType

        entity_id : str

        path : str

        mime_type : typing.Optional[str]

        project_name : typing.Optional[str]
            If null, the default project is used

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[StartMultipartUploadResponse]
            MultipartUploadResponse
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/attachment/upload-start",
            method="POST",
            json={
                "file_name": file_name,
                "num_of_file_parts": num_of_file_parts,
                "mime_type": mime_type,
                "project_name": project_name,
                "entity_type": entity_type,
                "entity_id": entity_id,
                "path": path,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    StartMultipartUploadResponse,
                    parse_obj_as(
                        type_=StartMultipartUploadResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def upload_attachment(
        self,
        *,
        file_name: str,
        entity_type: UploadAttachmentRequestEntityType,
        entity_id: str,
        request: typing.Dict[str, typing.Optional[typing.Any]],
        project_name: typing.Optional[str] = None,
        mime_type: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Upload attachment to MinIO

        Parameters
        ----------
        file_name : str

        entity_type : UploadAttachmentRequestEntityType

        entity_id : str

        request : typing.Dict[str, typing.Optional[typing.Any]]

        project_name : typing.Optional[str]

        mime_type : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/attachment/upload",
            method="PUT",
            params={
                "file_name": file_name,
                "project_name": project_name,
                "mime_type": mime_type,
                "entity_type": entity_type,
                "entity_id": entity_id,
            },
            json=request,
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .error_info import ErrorInfo
from .json_list_string import JsonListString
from .trace_update_source import TraceUpdateSource


class TraceUpdate(UniversalBaseModel):
    project_name: typing.Optional[str] = pydantic.Field(default=None)
    """
    If null and project_id not specified, Default Project is assumed
    """

    project_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    If null and project_name not specified, Default Project is assumed
    """

    name: typing.Optional[str] = None
    end_time: typing.Optional[dt.datetime] = None
    input: typing.Optional[JsonListString] = None
    output: typing.Optional[JsonListString] = None
    metadata: typing.Optional[JsonListString] = None
    tags: typing.Optional[typing.List[str]] = pydantic.Field(default=None)
    """
    Tags
    """

    tags_to_add: typing.Optional[typing.List[str]] = pydantic.Field(default=None)
    """
    Tags to add
    """

    tags_to_remove: typing.Optional[typing.List[str]] = pydantic.Field(default=None)
    """
    Tags to remove
    """

    error_info: typing.Optional[ErrorInfo] = None
    thread_id: typing.Optional[str] = None
    ttft: typing.Optional[float] = None
    source: typing.Optional[TraceUpdateSource] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



"""Migration manifest for tracking import state and enabling resumable migrations.

Uses a SQLite database (``migration_manifest.db``) for atomic, efficient,
crash-safe persistence.

Design
------
* **WAL mode** – SQLite's Write-Ahead Log allows safe, concurrent reads
  during writes and recovers cleanly from process crashes.
* **``synchronous=NORMAL``** – with WAL this is safe against process crashes
  (data survives) while being significantly faster than ``FULL``.
* **Batched writes** – ``mark_file_completed``, ``mark_file_failed``, and
  ``add_trace_mapping`` buffer changes in memory and commit every
  ``batch_size`` operations (default 50) in a single transaction.
  Lifecycle calls (``start``, ``complete``, ``reset``) always flush
  immediately so that import state is never stale on disk.
* **Idempotent writes** – ``INSERT OR IGNORE`` / ``INSERT OR REPLACE``
  mean re-running after a crash produces no duplicates in the DB.
* **In-memory deduplication** – pending buffers use ``set``/``dict`` so
  duplicate calls within a batch are collapsed before hitting SQLite.

Crash trade-off
---------------
Writes within a batch are buffered in memory. If the process is killed
between automatic flushes, up to ``batch_size - 1`` file completions may be
absent from the manifest on the next run. Those files will be re-imported
(idempotent for datasets/prompts; may create a duplicate trace — see the
project import docs). ``__del__`` provides a best-effort flush on normal
process exit. Use ``batch_size=1`` for per-file durability at the cost of
more disk transactions.

Lifecycle
---------
* No manifest → fresh import; ``MigrationManifest`` should only be
  constructed when a real import is about to start (not during ``--dry-run``).
* ``status = in_progress`` → interrupted; resume skips completed files.
* ``status = completed`` → nothing re-imported unless ``--force`` resets.
* ``--force`` → ``reset()`` wipes all tables and starts fresh.
"""

import logging
import os
import sqlite3
from datetime import datetime, timezone
from pathlib import Path
from typing import Dict, Set

LOGGER = logging.getLogger(__name__)

MANIFEST_FILENAME = "migration_manifest.db"


# Number of pending write operations that triggers an automatic flush to disk.
# Valid range: 1–10000 entries per SQLite transaction commit.
# Override with the OPIK_MIGRATION_BATCH_SIZE environment variable.
def _resolve_batch_size() -> int:
    raw = os.getenv("OPIK_MIGRATION_BATCH_SIZE", "")
    if raw:
        try:
            val = int(raw)
            if 1 <= val <= 10000:
                return val
            LOGGER.warning(
                "OPIK_MIGRATION_BATCH_SIZE=%r is outside the valid range 1–10000; "
                "using default of 50.",
                raw,
            )
        except ValueError:
            LOGGER.warning(
                "OPIK_MIGRATION_BATCH_SIZE=%r is not a valid integer; "
                "using default of 50.",
                raw,
            )
    return 50


DEFAULT_BATCH_SIZE: int = _resolve_batch_size()

_SCHEMA_SQL = """
CREATE TABLE IF NOT EXISTS status (
    key   TEXT PRIMARY KEY,
    value TEXT NOT NULL
);
CREATE TABLE IF NOT EXISTS completed_files (
    path TEXT PRIMARY KEY
);
CREATE TABLE IF NOT EXISTS failed_files (
    path  TEXT PRIMARY KEY,
    error TEXT NOT NULL
);
CREATE TABLE IF NOT EXISTS trace_id_map (
    src_id  TEXT PRIMARY KEY,
    dest_id TEXT NOT NULL
);
"""


class MigrationManifest:
    """Tracks import progress so interrupted migrations can be resumed safely.

    Parameters
    ----------
    base_path:
        Root directory of the export tree.  The ``migration_manifest.db``
        file is created here.
    batch_size:
        Number of pending write operations that triggers an automatic flush
        to disk.  Larger values mean fewer disk transactions (faster) but
        more work potentially lost if the process is killed between flushes.
        Defaults to ``DEFAULT_BATCH_SIZE`` (50).
    """

    def __init__(self, base_path: Path, batch_size: int = DEFAULT_BATCH_SIZE) -> None:
        self.base_path = base_path
        self.manifest_file = base_path / MANIFEST_FILENAME
        self._batch_size = batch_size

        # In-memory write buffers — flushed in one transaction per batch.
        # Using set/dict collapses duplicates before they reach SQLite.
        self._pending_completed: Set[str] = set()
        self._pending_failed: Dict[str, str] = {}  # path -> error
        self._pending_trace_map: Dict[str, str] = {}  # src_id -> dest_id

        self._conn = self._open()

    # ------------------------------------------------------------------
    # Construction helpers
    # ------------------------------------------------------------------

    def _open(self) -> sqlite3.Connection:
        self.base_path.mkdir(parents=True, exist_ok=True)
        conn = sqlite3.connect(
            str(self.manifest_file),
            # autocommit mode — we manage every transaction explicitly with
            # "with conn:" so there are no implicit BEGIN surprises.
            isolation_level=None,
            check_same_thread=True,
        )
        # WAL: persistent per-file setting, safe to set on every open.
        conn.execute("PRAGMA journal_mode=WAL")
        # NORMAL sync: per-connection setting, must be applied each open.
        conn.execute("PRAGMA synchronous=NORMAL")
        # Create tables if they don't exist yet.
        # executescript always commits first; fine here since we just opened.
        conn.executescript(_SCHEMA_SQL)
        return conn

    @classmethod
    def exists(cls, base_path: Path) -> bool:
        return (base_path / MANIFEST_FILENAME).exists()

    # ------------------------------------------------------------------
    # Status
    # ------------------------------------------------------------------

    @property
    def status(self) -> str:
        row = self._conn.execute(
            "SELECT value FROM status WHERE key = 'status'"
        ).fetchone()
        return row[0] if row else "not_started"

    @property
    def is_in_progress(self) -> bool:
        return self.status == "in_progress"

    @property
    def is_completed(self) -> bool:
        return self.status == "completed"

    def start(self) -> None:
        """Mark the import as in-progress (idempotent on resume)."""
        self._flush()
        with self._conn:
            # Preserve the original started_at across resume calls.
            self._conn.execute(
                "INSERT OR IGNORE INTO status(key, value) VALUES ('started_at', ?)",
                (datetime.now(timezone.utc).isoformat(),),
            )
            self._conn.execute(
                "INSERT OR REPLACE INTO status(key, value) VALUES ('status', 'in_progress')",
            )

    def complete(self) -> None:
        """Mark the import as fully completed, flushing any buffered writes."""
        self._flush()
        with self._conn:
            self._conn.execute(
                "INSERT OR REPLACE INTO status(key, value) VALUES ('status', 'completed')",
            )
            self._conn.execute(
                "INSERT OR REPLACE INTO status(key, value) VALUES ('completed_at', ?)",
                (datetime.now(timezone.utc).isoformat(),),
            )

    def reset(self) -> None:
        """Discard all tracked state and start fresh (used by ``--force``)."""
        # Drop buffered writes — they belong to the old run.
        self._pending_completed.clear()
        self._pending_failed.clear()
        self._pending_trace_map.clear()
        with self._conn:
            self._conn.execute("DELETE FROM status")
            self._conn.execute("DELETE FROM completed_files")
            self._conn.execute("DELETE FROM failed_files")
            self._conn.execute("DELETE FROM trace_id_map")

    # ------------------------------------------------------------------
    # File-level tracking
    # ------------------------------------------------------------------

    def relative_path(self, file_path: Path) -> str:
        """Return the canonical relative path string used as a manifest key."""
        return str(file_path.relative_to(self.base_path))

    def is_file_completed(self, file_path: Path) -> bool:
        rel = self.relative_path(file_path)
        # Check the in-memory buffer first (O(1)); avoids a DB round-trip for
        # files that were completed in the current batch but not yet flushed.
        if rel in self._pending_completed:
            return True
        row = self._conn.execute(
            "SELECT 1 FROM completed_files WHERE path = ?", (rel,)
        ).fetchone()
        return row is not None

    def mark_file_completed(self, file_path: Path) -> None:
        rel = self.relative_path(file_path)
        self._pending_completed.add(rel)
        # Completion supersedes any earlier failure recorded in the buffer.
        self._pending_failed.pop(rel, None)
        self._maybe_flush()

    def mark_file_failed(self, file_path: Path, error: str) -> None:
        rel = self.relative_path(file_path)
        self._pending_failed[rel] = error
        self._maybe_flush()

    # ------------------------------------------------------------------
    # Trace ID mapping (src_id -> dest_id)
    # Persisted so experiment cross-references survive resume across sessions.
    # ------------------------------------------------------------------

    def add_trace_mapping(self, src_id: str, dest_id: str) -> None:
        self._pending_trace_map[src_id] = dest_id
        self._maybe_flush()

    def get_trace_id_map(self) -> Dict[str, str]:
        """Return the full trace ID mapping including any unflushed entries."""
        self._flush()
        rows = self._conn.execute("SELECT src_id, dest_id FROM trace_id_map").fetchall()
        return {src: dest for src, dest in rows}

    # ------------------------------------------------------------------
    # Public flush
    # ------------------------------------------------------------------

    def save(self) -> None:
        """Flush any buffered writes to disk immediately."""
        self._flush()

    # ------------------------------------------------------------------
    # Stats
    # ------------------------------------------------------------------

    def completed_count(self) -> int:
        self._flush()
        row = self._conn.execute("SELECT COUNT(*) FROM completed_files").fetchone()
        return row[0] if row else 0

    def failed_count(self) -> int:
        self._flush()
        row = self._conn.execute("SELECT COUNT(*) FROM failed_files").fetchone()
        return row[0] if row else 0

    def get_failed_files(self) -> Dict[str, str]:
        self._flush()
        rows = self._conn.execute("SELECT path, error FROM failed_files").fetchall()
        return {path: error for path, error in rows}

    # ------------------------------------------------------------------
    # Lifecycle
    # ------------------------------------------------------------------

    def close(self) -> None:
        """Flush pending writes and release the DB connection."""
        try:
            self._flush()
        finally:
            self._conn.close()

    def __del__(self) -> None:
        """Best-effort flush when the object is garbage-collected."""
        try:
            self._flush()
        except Exception:
            pass

    # ------------------------------------------------------------------
    # Internal
    # ------------------------------------------------------------------

    @property
    def _pending_count(self) -> int:
        return (
            len(self._pending_completed)
            + len(self._pending_failed)
            + len(self._pending_trace_map)
        )

    def _maybe_flush(self) -> None:
        if self._pending_count >= self._batch_size:
            self._flush()

    def _flush(self) -> None:
        """Commit all buffered writes in a single atomic transaction."""
        if self._pending_count == 0:
            return

        with self._conn:
            if self._pending_completed:
                self._conn.executemany(
                    "INSERT OR IGNORE INTO completed_files(path) VALUES (?)",
                    [(p,) for p in self._pending_completed],
                )
                # Remove from failed_files if a previous run recorded a failure.
                self._conn.executemany(
                    "DELETE FROM failed_files WHERE path = ?",
                    [(p,) for p in self._pending_completed],
                )
                self._pending_completed.clear()

            if self._pending_failed:
                self._conn.executemany(
                    "INSERT OR REPLACE INTO failed_files(path, error) VALUES (?, ?)",
                    list(self._pending_failed.items()),
                )
                self._pending_failed.clear()

            if self._pending_trace_map:
                self._conn.executemany(
                    "INSERT OR REPLACE INTO trace_id_map(src_id, dest_id) VALUES (?, ?)",
                    list(self._pending_trace_map.items()),
                )
                self._pending_trace_map.clear()



"""Dataset import functionality."""

import json
from pathlib import Path
from typing import Dict, Optional

import opik
from rich.console import Console

from ..migration_manifest import MigrationManifest
from .utils import matches_name_pattern

console = Console()


def import_datasets_from_directory(
    client: opik.Opik,
    source_dir: Path,
    dry_run: bool,
    name_pattern: Optional[str],
    debug: bool,
    manifest: Optional[MigrationManifest] = None,
) -> Dict[str, int]:
    """Import datasets from a directory.

    Returns:
        Dictionary with keys: 'datasets', 'datasets_skipped', 'datasets_errors'
    """
    try:
        dataset_files = list(source_dir.glob("dataset_*.json"))

        if not dataset_files:
            console.print("[yellow]No dataset files found in the directory[/yellow]")
            return {"datasets": 0, "datasets_skipped": 0, "datasets_errors": 0}

        imported_count = 0
        skipped_count = 0
        error_count = 0
        for dataset_file in dataset_files:
            try:
                # Skip files already completed in a previous run
                if (
                    manifest
                    and not dry_run
                    and manifest.is_file_completed(dataset_file)
                ):
                    if debug:
                        console.print(
                            f"[blue]Skipping {dataset_file.name} (already imported in a previous run)[/blue]"
                        )
                    skipped_count += 1
                    continue

                with open(dataset_file, "r", encoding="utf-8") as f:
                    dataset_data = json.load(f)

                # Handle two export formats:
                # 1. {"name": "...", "items": [...]} - from export_single_dataset
                # 2. {"dataset": {"name": "...", "id": "..."}, "items": [...]} - from export_experiment_datasets
                dataset_name = dataset_data.get("name") or (
                    dataset_data.get("dataset", {}).get("name")
                    if dataset_data.get("dataset")
                    else None
                )

                # Check if name is missing or empty
                if not dataset_name or (
                    isinstance(dataset_name, str) and not dataset_name.strip()
                ):
                    console.print(
                        f"[red]Error: Dataset file {dataset_file.name} is missing or has an empty name field[/red]"
                    )
                    error_count += 1
                    continue

                # Strip whitespace from name
                dataset_name = dataset_name.strip()

                # Filter by name pattern if specified
                if name_pattern and not matches_name_pattern(
                    dataset_name, name_pattern
                ):
                    if debug:
                        console.print(
                            f"[blue]Skipping dataset {dataset_name} (doesn't match pattern)[/blue]"
                        )
                    skipped_count += 1
                    continue

                if dry_run:
                    console.print(f"[blue]Would import dataset: {dataset_name}[/blue]")
                    imported_count += 1
                    continue

                if debug:
                    console.print(f"[blue]Importing dataset: {dataset_name}[/blue]")

                # Get or create dataset (handles case where dataset already exists)
                try:
                    dataset = client.get_dataset(dataset_name)
                    if debug:
                        console.print(
                            f"[blue]Dataset '{dataset_name}' already exists, using existing dataset[/blue]"
                        )
                except Exception:
                    # Dataset doesn't exist, create it
                    dataset = client.create_dataset(name=dataset_name)
                    if debug:
                        console.print(
                            f"[blue]Created new dataset: {dataset_name}[/blue]"
                        )

                # Import dataset items
                items = dataset_data.get("items", [])
                if items:
                    # Remove 'id' field from items before inserting (IDs are auto-generated)
                    items_to_insert = []
                    for item in items:
                        if isinstance(item, dict):
                            # Create a copy without the 'id' field
                            item_copy = {k: v for k, v in item.items() if k != "id"}
                            items_to_insert.append(item_copy)
                        else:
                            items_to_insert.append(item)

                    if items_to_insert:
                        dataset.insert(items_to_insert)
                        if debug:
                            console.print(
                                f"[blue]Inserted {len(items_to_insert)} items into dataset '{dataset_name}'[/blue]"
                            )
                    else:
                        console.print(
                            f"[yellow]Warning: No items to insert for dataset '{dataset_name}' (all items were empty after removing 'id' field)[/yellow]"
                        )

                imported_count += 1
                if debug:
                    console.print(
                        f"[green]Imported dataset: {dataset_name} with {len(items)} items[/green]"
                    )

                # Mark this file as completed in the manifest
                if manifest:
                    manifest.mark_file_completed(dataset_file)

            except Exception as e:
                console.print(
                    f"[red]Error importing dataset from {dataset_file.name}: {e}[/red]"
                )
                if manifest:
                    manifest.mark_file_failed(dataset_file, str(e))
                error_count += 1
                continue

        return {
            "datasets": imported_count,
            "datasets_skipped": skipped_count,
            "datasets_errors": error_count,
        }

    except Exception as e:
        console.print(f"[red]Error importing datasets: {e}[/red]")
        return {"datasets": 0, "datasets_skipped": 0, "datasets_errors": 1}



"""Constants for usage report module."""

# Safety limit to avoid infinite loops in pagination
MAX_PAGINATION_PAGES = 1000

# Maximum number of trace results to fetch per project when counting spans.
# This limit prevents excessive API calls for projects with very large trace counts.
# If a project has more traces than this limit, span counts may be incomplete.
MAX_TRACE_RESULTS = 10000



"""In-memory entrypoint registry for the local runner."""

import dataclasses
import inspect
import threading
from typing import Any, Callable, Dict, List

_lock = threading.Lock()
REGISTRY: Dict[str, Dict[str, Any]] = {}
_listeners: List[Callable[[str], None]] = []


@dataclasses.dataclass
class Param:
    name: str
    type: str = "str"


def register(
    name: str,
    func: Callable,
    project: str,
    params: List[Param],
    docstring: str,
) -> None:
    with _lock:
        REGISTRY[name] = {
            "func": func,
            "name": name,
            "project": project,
            "params": params,
            "docstring": docstring,
        }
        listeners = list(_listeners)

    for listener in listeners:
        listener(name)


def on_register(listener: Callable[[str], None]) -> None:
    with _lock:
        _listeners.append(listener)


def get_all() -> Dict[str, Dict[str, Any]]:
    with _lock:
        return dict(REGISTRY)


def extract_params(fn: Callable) -> List[Param]:
    sig = inspect.signature(fn)
    params: List[Param] = []
    for param_name, param in sig.parameters.items():
        if param.annotation is inspect.Parameter.empty:
            type_name = "str"
        else:
            ann = param.annotation
            type_name = ann.__name__ if hasattr(ann, "__name__") else str(ann)
        params.append(Param(name=param_name, type=type_name))
    return params



import logging
import time
from typing import Any, Dict, Optional

import httpx

from ..api_objects import opik_client, rest_helpers
from ..rest_api import client as rest_api_client

LOGGER = logging.getLogger(__name__)

DEFAULT_TIMEOUT_SECONDS = 120
DEFAULT_POLL_INTERVAL_SECONDS = 0.5


class LocalRunnerTask:
    """
    A callable task that submits jobs to the local runner and polls for completion.
    Conforms to the ``LLMTask`` protocol (``Callable[[Dict[str, Any]], Dict[str, Any]]``)
    for use with ``evaluate()`` and ``EvaluationSuite``.

    When called, the instance:

    1. Filters the ``"id"`` key from the input dict (dataset-item artefact).
    2. Submits a job to the local runner via the REST API.
    3. Polls until the job reaches a terminal status or the timeout expires.
    4. Returns ``{"input": <original item>, "output": <job result>}``.

    Failures raise ``RuntimeError``; timeouts raise ``TimeoutError``.

    Args:
        project_name: The project to submit jobs under.
        agent_name: The agent to execute the job.
        mask_id: Optional agent config override mask.
        timeout_seconds: Max time to wait for job completion (default 120).
        poll_interval_seconds: Interval between status polls (default 0.5).
    """

    def __init__(
        self,
        project_name: str,
        agent_name: str,
        mask_id: Optional[str] = None,
        timeout_seconds: float = DEFAULT_TIMEOUT_SECONDS,
        poll_interval_seconds: float = DEFAULT_POLL_INTERVAL_SECONDS,
    ) -> None:
        if poll_interval_seconds <= 0:
            raise ValueError("poll_interval_seconds must be positive")

        self._project_name = project_name
        self._agent_name = agent_name
        self._mask_id = mask_id
        self._timeout_seconds = timeout_seconds
        self._poll_interval_seconds = poll_interval_seconds
        self._project_id: Optional[str] = None

    def _get_rest_client(self) -> rest_api_client.OpikApi:
        return opik_client.get_client_cached().rest_client

    def _resolve_project_id(self) -> str:
        if self._project_id is None:
            self._project_id = rest_helpers.resolve_project_id_by_name(
                self._get_rest_client(), self._project_name
            )
        return self._project_id

    def _submit_job(self, inputs: Dict[str, Any]) -> str:
        rest_client = self._get_rest_client()
        project_id = self._resolve_project_id()

        kwargs: Dict[str, Any] = {
            "agent_name": self._agent_name,
            "project_id": project_id,
            "inputs": inputs,
        }
        if self._mask_id is not None:
            kwargs["mask_id"] = self._mask_id

        try:
            # create_job raises for non-2xx responses internally
            resp = rest_client.runners.with_raw_response.create_job(**kwargs)
        except (httpx.ConnectError, httpx.TimeoutException) as exc:
            raise RuntimeError(
                f"Failed to submit job due to network error: {exc}"
            ) from exc

        job_id = resp.headers.get("location", "").rsplit("/", 1)[-1]
        if not job_id:
            raise RuntimeError("Could not extract job ID from Location header")
        return job_id

    def _wait_for_job(self, job_id: str) -> Dict[str, Any]:
        rest_client = self._get_rest_client()
        deadline = time.monotonic() + self._timeout_seconds

        while True:
            try:
                job = rest_client.runners.get_job(job_id)
            except (httpx.ConnectError, httpx.TimeoutException) as exc:
                LOGGER.warning(
                    "Transient network error polling job %s: %s", job_id, exc
                )
                if time.monotonic() >= deadline:
                    raise TimeoutError(
                        f"Job {job_id} did not complete within {self._timeout_seconds}s"
                    ) from exc
                time.sleep(self._poll_interval_seconds)
                continue

            if job.status == "completed":
                return dict(job.result) if job.result else {}
            if job.status == "failed":
                raise RuntimeError(f"Job {job_id} failed: {job.error}")
            if job.status == "cancelled":
                raise RuntimeError(f"Job {job_id} was cancelled")

            if time.monotonic() >= deadline:
                raise TimeoutError(
                    f"Job {job_id} did not complete within {self._timeout_seconds}s"
                )
            time.sleep(self._poll_interval_seconds)

    def __call__(self, item: Dict[str, Any]) -> Dict[str, Any]:
        # Filter out "id" — it's a dataset-item identifier, not an agent input
        inputs = {k: v for k, v in item.items() if k != "id"}
        result = self._wait_for_job(self._submit_job(inputs))
        output = result.get("result", result)
        return {
            "input": item,
            "output": output,
        }



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .provider_api_key_write_provider import ProviderApiKeyWriteProvider

__all__ = ["ProviderApiKeyWriteProvider"]



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.manual_evaluation_request_entity_type import ManualEvaluationRequestEntityType
from ..types.manual_evaluation_response import ManualEvaluationResponse
from .raw_client import AsyncRawManualEvaluationClient, RawManualEvaluationClient

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class ManualEvaluationClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawManualEvaluationClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawManualEvaluationClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawManualEvaluationClient
        """
        return self._raw_client

    def evaluate_spans(
        self,
        *,
        project_id: str,
        entity_ids: typing.Sequence[str],
        rule_ids: typing.Sequence[str],
        entity_type: ManualEvaluationRequestEntityType,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ManualEvaluationResponse:
        """
        Manually trigger evaluation rules on selected spans. Bypasses sampling and enqueues all specified spans for evaluation.

        Parameters
        ----------
        project_id : str
            Project ID

        entity_ids : typing.Sequence[str]
            List of entity IDs (trace IDs or thread IDs) to evaluate

        rule_ids : typing.Sequence[str]
            List of automation rule IDs to apply

        entity_type : ManualEvaluationRequestEntityType
            Type of entity to evaluate (trace or thread)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ManualEvaluationResponse
            Accepted - Evaluation request queued successfully

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.manual_evaluation.evaluate_spans(project_id='550e8400-e29b-41d4-a716-446655440000', entity_ids=['550e8400-e29b-41d4-a716-446655440000', '550e8400-e29b-41d4-a716-446655440001'], rule_ids=['660e8400-e29b-41d4-a716-446655440000'], entity_type="trace", )
        """
        _response = self._raw_client.evaluate_spans(
            project_id=project_id,
            entity_ids=entity_ids,
            rule_ids=rule_ids,
            entity_type=entity_type,
            request_options=request_options,
        )
        return _response.data

    def evaluate_threads(
        self,
        *,
        project_id: str,
        entity_ids: typing.Sequence[str],
        rule_ids: typing.Sequence[str],
        entity_type: ManualEvaluationRequestEntityType,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ManualEvaluationResponse:
        """
        Manually trigger evaluation rules on selected threads. Bypasses sampling and enqueues all specified threads for evaluation.

        Parameters
        ----------
        project_id : str
            Project ID

        entity_ids : typing.Sequence[str]
            List of entity IDs (trace IDs or thread IDs) to evaluate

        rule_ids : typing.Sequence[str]
            List of automation rule IDs to apply

        entity_type : ManualEvaluationRequestEntityType
            Type of entity to evaluate (trace or thread)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ManualEvaluationResponse
            Accepted - Evaluation request queued successfully

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.manual_evaluation.evaluate_threads(project_id='550e8400-e29b-41d4-a716-446655440000', entity_ids=['550e8400-e29b-41d4-a716-446655440000', '550e8400-e29b-41d4-a716-446655440001'], rule_ids=['660e8400-e29b-41d4-a716-446655440000'], entity_type="trace", )
        """
        _response = self._raw_client.evaluate_threads(
            project_id=project_id,
            entity_ids=entity_ids,
            rule_ids=rule_ids,
            entity_type=entity_type,
            request_options=request_options,
        )
        return _response.data

    def evaluate_traces(
        self,
        *,
        project_id: str,
        entity_ids: typing.Sequence[str],
        rule_ids: typing.Sequence[str],
        entity_type: ManualEvaluationRequestEntityType,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ManualEvaluationResponse:
        """
        Manually trigger evaluation rules on selected traces. Bypasses sampling and enqueues all specified traces for evaluation.

        Parameters
        ----------
        project_id : str
            Project ID

        entity_ids : typing.Sequence[str]
            List of entity IDs (trace IDs or thread IDs) to evaluate

        rule_ids : typing.Sequence[str]
            List of automation rule IDs to apply

        entity_type : ManualEvaluationRequestEntityType
            Type of entity to evaluate (trace or thread)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ManualEvaluationResponse
            Accepted - Evaluation request queued successfully

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.manual_evaluation.evaluate_traces(project_id='550e8400-e29b-41d4-a716-446655440000', entity_ids=['550e8400-e29b-41d4-a716-446655440000', '550e8400-e29b-41d4-a716-446655440001'], rule_ids=['660e8400-e29b-41d4-a716-446655440000'], entity_type="trace", )
        """
        _response = self._raw_client.evaluate_traces(
            project_id=project_id,
            entity_ids=entity_ids,
            rule_ids=rule_ids,
            entity_type=entity_type,
            request_options=request_options,
        )
        return _response.data


class AsyncManualEvaluationClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawManualEvaluationClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawManualEvaluationClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawManualEvaluationClient
        """
        return self._raw_client

    async def evaluate_spans(
        self,
        *,
        project_id: str,
        entity_ids: typing.Sequence[str],
        rule_ids: typing.Sequence[str],
        entity_type: ManualEvaluationRequestEntityType,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ManualEvaluationResponse:
        """
        Manually trigger evaluation rules on selected spans. Bypasses sampling and enqueues all specified spans for evaluation.

        Parameters
        ----------
        project_id : str
            Project ID

        entity_ids : typing.Sequence[str]
            List of entity IDs (trace IDs or thread IDs) to evaluate

        rule_ids : typing.Sequence[str]
            List of automation rule IDs to apply

        entity_type : ManualEvaluationRequestEntityType
            Type of entity to evaluate (trace or thread)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ManualEvaluationResponse
            Accepted - Evaluation request queued successfully

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.manual_evaluation.evaluate_spans(project_id='550e8400-e29b-41d4-a716-446655440000', entity_ids=['550e8400-e29b-41d4-a716-446655440000', '550e8400-e29b-41d4-a716-446655440001'], rule_ids=['660e8400-e29b-41d4-a716-446655440000'], entity_type="trace", )
        asyncio.run(main())
        """
        _response = await self._raw_client.evaluate_spans(
            project_id=project_id,
            entity_ids=entity_ids,
            rule_ids=rule_ids,
            entity_type=entity_type,
            request_options=request_options,
        )
        return _response.data

    async def evaluate_threads(
        self,
        *,
        project_id: str,
        entity_ids: typing.Sequence[str],
        rule_ids: typing.Sequence[str],
        entity_type: ManualEvaluationRequestEntityType,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ManualEvaluationResponse:
        """
        Manually trigger evaluation rules on selected threads. Bypasses sampling and enqueues all specified threads for evaluation.

        Parameters
        ----------
        project_id : str
            Project ID

        entity_ids : typing.Sequence[str]
            List of entity IDs (trace IDs or thread IDs) to evaluate

        rule_ids : typing.Sequence[str]
            List of automation rule IDs to apply

        entity_type : ManualEvaluationRequestEntityType
            Type of entity to evaluate (trace or thread)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ManualEvaluationResponse
            Accepted - Evaluation request queued successfully

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.manual_evaluation.evaluate_threads(project_id='550e8400-e29b-41d4-a716-446655440000', entity_ids=['550e8400-e29b-41d4-a716-446655440000', '550e8400-e29b-41d4-a716-446655440001'], rule_ids=['660e8400-e29b-41d4-a716-446655440000'], entity_type="trace", )
        asyncio.run(main())
        """
        _response = await self._raw_client.evaluate_threads(
            project_id=project_id,
            entity_ids=entity_ids,
            rule_ids=rule_ids,
            entity_type=entity_type,
            request_options=request_options,
        )
        return _response.data

    async def evaluate_traces(
        self,
        *,
        project_id: str,
        entity_ids: typing.Sequence[str],
        rule_ids: typing.Sequence[str],
        entity_type: ManualEvaluationRequestEntityType,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ManualEvaluationResponse:
        """
        Manually trigger evaluation rules on selected traces. Bypasses sampling and enqueues all specified traces for evaluation.

        Parameters
        ----------
        project_id : str
            Project ID

        entity_ids : typing.Sequence[str]
            List of entity IDs (trace IDs or thread IDs) to evaluate

        rule_ids : typing.Sequence[str]
            List of automation rule IDs to apply

        entity_type : ManualEvaluationRequestEntityType
            Type of entity to evaluate (trace or thread)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ManualEvaluationResponse
            Accepted - Evaluation request queued successfully

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.manual_evaluation.evaluate_traces(project_id='550e8400-e29b-41d4-a716-446655440000', entity_ids=['550e8400-e29b-41d4-a716-446655440000', '550e8400-e29b-41d4-a716-446655440001'], rule_ids=['660e8400-e29b-41d4-a716-446655440000'], entity_type="trace", )
        asyncio.run(main())
        """
        _response = await self._raw_client.evaluate_traces(
            project_id=project_id,
            entity_ids=entity_ids,
            rule_ids=rule_ids,
            entity_type=entity_type,
            request_options=request_options,
        )
        return _response.data
